@@ -32,17 +32,29 @@ namespace GtkSharp.Generation {
 		Parameters parms;
 		bool is_static = false;
 		string mods = String.Empty;
+		string name;
 
 		protected MethodBase (XmlElement elem, ClassBase container_type) 
 		{
 			this.elem = elem;
 			this.container_type = container_type;
+			this.name = elem.GetAttribute ("name");
 			parms = new Parameters (elem ["parameters"]);
 			IsStatic = elem.GetAttribute ("shared") == "true";
 			if (elem.HasAttribute ("new_flag"))
 				mods = "new ";
 		}
 
+		protected string BaseName {
+			get {
+				string name = Name;
+				int idx = Name.LastIndexOf (".");
+				if (idx > 0)
+					name = Name.Substring (idx + 1);
+				return name;
+			}
+		}
+
 		MethodBody body;
 		public MethodBody Body {
 			get {
@@ -58,12 +70,27 @@ namespace GtkSharp.Generation {
 			}
 		}
 
-		ImportSignature isig;
-		public ImportSignature ImportSignature {
+		protected bool HasGetterName {
 			get {
-				if (isig == null)
-					isig = new ImportSignature (parms);
-				return isig;
+				string name = BaseName;
+				if (name.Length <= 3)
+					return false;
+				if (name.StartsWith ("Get") || name.StartsWith ("Has"))
+					return Char.IsUpper (name [3]);
+				else if (name.StartsWith ("Is"))
+					return Char.IsUpper (name [2]);
+				else
+					return false;
+			}
+		}
+
+		protected bool HasSetterName {
+			get {
+				string name = BaseName;
+				if (name.Length <= 3)
+					return false;
+
+				return name.StartsWith ("Set") && Char.IsUpper (name [3]);
 			}
 		}
 
@@ -94,6 +121,15 @@ namespace GtkSharp.Generation {
 			}
 		}
 
+		public string Name {
+			get {
+				return name;
+			}
+			set {
+				name = value;
+			}
+		}
+
 		public Parameters Parameters {
 			get {
 				return parms;
